from commandlib import Command, CommandError
from path import Path
import patoolib
import shutil
import os


def log(message):
    print(message)


def extract_archive(filename, directory):
    patoolib.extract_archive(filename, outdir=directory)


class DownloadError(Exception):
    pass


def download_file(downloaded_file_path, url, max_connections=2, max_concurrent=5):
    """Download file to specified location."""
    file_path = Path(downloaded_file_path)
    assert file_path.isabs(), "download file path must be absolute, not relative"
    if file_path.exists():
        log("{} already downloaded".format(file_path))
        return

    log("Downloading: {}\n".format(url))
    aria2c = Command("aria2c")
    aria2c = aria2c("--max-connection-per-server={}".format(max_connections))
    aria2c = aria2c("--max-concurrent-downloads={}".format(max_concurrent))

    try:
        aria2c(
            "--dir={}".format(file_path.dirname()),
            "--out={}.part".format(file_path.basename()),
            url
        ).run()
    except CommandError:
        raise DownloadError(
            "Failed to download {}. Re-running may fix the problem.".format(url)
        )

    shutil.move(file_path + ".part", file_path)



from hitchstory import InfoDefinition, InfoProperty
from strictyaml import Seq, Enum
import hitchpylibrarytoolkit
import hitchchrome
import os


class Engine(hitchpylibrarytoolkit.Engine):
    info_definition = InfoDefinition(
        environments=InfoProperty(Seq(Enum([
            "gui", "mac", "docker", "headless", "wsl"
        ]))),
    )
        
    def set_up(self):
        super().set_up()
        self._chrome = hitchchrome.ChromeBuild(
            self._build._paths.gen / "devchrome"
        )
        self._chrome.ensure_built()
        if "EXTERNAL_CHROME" not in os.environ:
            os.environ["EXTERNAL_CHROME"] = str(self._chrome.chrome_bin)

    def screenshot_exists(self, filename):
        assert self._build.working.joinpath(filename).exists()



from hitchstory import HitchStoryException, StoryCollection
from hitchrun import expected
from commandlib import CommandError
from strictyaml import Str, Map, Bool, load
from pathquery import pathquery
from hitchrun import DIR
import dirtemplate
import hitchpylibrarytoolkit
from path import Path
from versionbullshit import get_versions
from engine import Engine
import json


toolkit = hitchpylibrarytoolkit.ProjectToolkit(
    "hitchchrome",
    DIR,
)


@expected(HitchStoryException)
def bdd(*keywords):
    """Run single story."""
    toolkit.bdd(Engine(toolkit.build), keywords)
    

@expected(HitchStoryException)
def regression():
    """Run all stories."""
    clean()
    toolkit.regression(Engine(toolkit.build))


@expected(CommandError)
def clean():
    """Clean out built chrome and temp directory"""
    DIR.gen.joinpath("chrome").rmtree(ignore_errors=True)
    DIR.gen.joinpath("tmp").rmtree(ignore_errors=True)


def deploy(version):
    """
    Deploy to pypi as specified version.
    """
    toolkit.deploy(version)


def checkversioner():
    """
    Check version getting works.
    """
    print(json.dumps(get_versions(), indent=4))


def writeversions():
    """
    Write new versions JSON.
    """
    versionsjson = DIR.project.joinpath("hitchchrome", "versions.json")
    versionsjson.write_text(json.dumps(get_versions(), indent=4))
    print(versionsjson.text())
    



from selenium.webdriver.chrome.options import Options
from selenium import webdriver
from commandlib import Command
from hitchchrome import utils
from path import Path
import hitchbuild
import stat
import json
import sys

VERSIONS_PATH = Path(__file__).dirname().abspath() / "versions.json"

class HitchChromeBuildException(Exception):
    pass


class ChromeBuild(hitchbuild.HitchBuild):
    def __init__(self, path, version=None):
        self._download_urls = json.loads(VERSIONS_PATH.text())
        self._versions_available = list(self._download_urls.keys())

        if version is None:
            version = self._versions_available[0]

        if version not in self._versions_available:
            raise HitchChromeBuildException((
                "Chrome version {} not available in this package, "
                "available versions are {}".format(
                    version,
                    ", ".join(self._versions_available),
                )
            ))

        self.buildpath = Path(path).abspath()
        self.version = version
        self._version = self.variable("version", version)
        self._package_version = self.variable("package_version", 1)
        self.fingerprint_path = self.buildpath / "fingerprint.txt"
    
    @property
    def chrome_bin(self):
        if self.os_name == "linux":
            return Command(
                self.buildpath / "chrome-linux" / "chrome"
            )
        else:
            return Command(
                self.buildpath / "chrome-mac" / "Chromium.app" / "Contents" / "MacOS" / "Chromium"
            )

    @property
    def chromedriver_bin(self):
        if self.os_name == "linux":
            return Command(self.buildpath / "chromedriver_linux64" / "chromedriver")
        else:
            return Command(self.buildpath / "chromedriver_mac64" / "chromedriver")
    
    def clean(self):
        self.buildpath.rmtree(ignore_errors=True)
    
    @property
    def os_name(self):
        if sys.platform == "linux2" or sys.platform == "linux":
            return "linux"
        elif sys.platform == "darwin":
            return "mac"
        else:
            raise Exception("Platform {} not supported :(".format(sys.platform))

    def should_rebuild(self):
        return (
            self.incomplete() or
            self._version.changed or
            self._package_version.changed
        )

    def ensure_built(self):
        if self.should_rebuild():
            if self.buildpath.exists():
                self.buildpath.rmtree()
            self.buildpath.mkdir()
            
            download_urls = self._download_urls[self.version]
            
            if self.os_name == "linux":
                chrome_download_url = download_urls["linux_chrome"]
                chromedriver_download_url = download_urls["linux_chromedriver"]
            elif self.os_name == "mac":
                chrome_download_url = download_urls["mac_chrome"]
                chromedriver_download_url = download_urls["mac_chromedriver"]

            download_to = self.tmp / "chrome-{}.zip".format(self.version)
            utils.download_file(download_to, chrome_download_url)

            if self.os_name == "mac":
                # patool has a bug on mac that breaks chromium
                Command("unzip", download_to).in_dir(self.buildpath).run()
            else:
                utils.extract_archive(download_to, self.buildpath)

            download_to.remove()

            chrome_bin = Path(self.chrome_bin)
            chrome_bin.chmod(
                chrome_bin.stat().st_mode | stat.S_IEXEC
            )

            # Install chromedriver
            download_to = self.tmp / "chromedriver-{}.zip".format(self.version)
            utils.download_file(download_to, chromedriver_download_url)
            utils.extract_archive(download_to, self.buildpath)
            download_to.remove()
            
            chromedriver_bin = Path(self.chromedriver_bin)
            chromedriver_bin.chmod(
                chromedriver_bin.stat().st_mode | stat.S_IEXEC
            )
            
            self.verify()
            self.refingerprint()
    
    def verify(self):
        assert self.version in self.chromedriver_bin("--version").output()
    
    def webdriver(self, headless=False, arguments=None):
        options = Options()
        options.binary_location = str(self.chrome_bin)
        options.headless = headless

        if arguments is not None:
            assert isinstance(arguments, list), 'arguments must be a list - e.g. ["--nosandbox"]'
            for argument in arguments:
                options.add_argument(argument)

        return webdriver.Chrome(
            options=options,
            executable_path=str(self.chromedriver_bin),
        )
                                      



import requests
import time

OMAHAPROX = "omahaprox"


def history(ossmall):
    return requests.get(
        f"https://{OMAHAPROX}y.appspot.com/history.json?channel=stable&os={ossmall}"
    ).json()

def get_for_os(which_version, ossmall, oslarge):
    print(f"Downloading {which_version} version data for {ossmall}...")
    time.sleep(2)
    historyjson = history(ossmall)
    version = historyjson[which_version]['version']
    major_version = version.split(".")[0]
    
    existed_positions = requests.get((
        "https://www.googleapis.com/storage/v1/b/chromium-browser-snapshots/o?delimiter=/&prefix={oslarge}/"
        "&fields=items(kind,mediaLink,metadata,name,size,updated),kind,prefixes,nextPageToken"
    )).json()
    
    base_position = requests.get(
        "https://{}y.appspot.com/deps.json?version={}".format(OMAHAPROX, version)
    ).json()['chromium_base_position']                                       

    download_urls_available = requests.get((
        "https://www.googleapis.com/storage/v1/b/chromium-browser-snapshots/o?"
        "delimiter=/&prefix={}/{}&"
        "fields=items(kind,mediaLink,metadata,name,size,updated),kind,prefixes,nextPageToken"
    ).format(oslarge, base_position[:-2])).json()['prefixes']
    
    assert len(download_urls_available) > 0
    
    prefixes = [x.split("/")[1] for x in download_urls_available]

    take_closest = lambda num,collection:min(collection, key = lambda x: abs(x-num))
    prefix = str(take_closest(int(base_position), [int(x) for x in prefixes]))
    
    #prefix = download_urls_available[-1].split("/")[1]
    
    download_urls = requests.get((
        "https://www.googleapis.com/storage/v1/b/chromium-browser-snapshots/o?"
        "delimiter=/&prefix={}/{}/&"
        "fields=items(kind,mediaLink,metadata,name,size,updated),kind,prefixes,nextPageToken"
    ).format(oslarge, prefix)).json()['items']
    
    chrome_download_url = None
    chromedriver_url = None
    
    for download_url_metadata in download_urls:
        if f"chrome-{ossmall}.zip" in download_url_metadata['mediaLink']:
            chrome_download_url = download_url_metadata['mediaLink']
            
        if "chromedriver" in download_url_metadata['mediaLink']:
            chromedriver_url = download_url_metadata['mediaLink']
    
    return major_version, chrome_download_url, chromedriver_url

def indexes_of_wanted_version_numbers(ossmall):
    indexes = []
    wanted_major_version_numbers = []
    all_version_numbers = [hjson['version'] for hjson in history(ossmall)]
    for index, version_number in enumerate(all_version_numbers):
        major_version = version_number.split(".")[0]
        if major_version not in wanted_major_version_numbers:
            if int(major_version) >= 80:
                wanted_major_version_numbers.append(major_version)
                indexes.append(index)
    return indexes

def get_versions():
    print("Downloading version data...")
    
    versions = {}
    
    for index in indexes_of_wanted_version_numbers("linux"):
        version, chrome_url, driver_url = get_for_os(index, "linux", "Linux_x64")
        
        if version not in versions:
            versions[version] = {}
        versions[version]['linux_chrome'] = chrome_url
        versions[version]['linux_chromedriver'] = driver_url
    
    for index in indexes_of_wanted_version_numbers("mac"):
        version, mac_chrome_url, mac_driver_url = get_for_os(index, "mac", "Mac")
        if version not in versions:
            versions[version] = {}
        versions[version]['mac_chrome'] = mac_chrome_url
        versions[version]['mac_chromedriver'] = mac_driver_url

    return versions
